/** 
 Author: Michael Sparks - sparksmb@gmail.com
 **/

var app = (function () {
  "use strict";
  /*global window, XMLHttpRequest, ActiveXObject */
  
	var engine,
      document = window.document;
  
	/** private methods **/
  function isNumber(value) {
    return !isNaN(Number(value));
  }
  
  function getParentObjOfViewArray(pathTokens) {
    var i, 
        objRef = app,  //the path root is always the app object thus this references the global app var
        objPathLen = pathTokens.length - 1; //exclude last token because it's the name of the view
   
    for (i = 0; i < objPathLen; i += 1) {
      objRef = objRef[pathTokens[i]];
    }
    return objRef;
  }
  
  function getViewIndexByViewName(viewArray, viewName) {
    var i,
        length = viewArray.length || 0,
        view;
    for (i = 0; i < length; i += 1) {
      view = viewArray[i];
      if (view.name === viewName) {
        return i;
      }
    }
    
    return -1;
  }
  
  function getViewIndex(views, viewName) {
    var index;
    
    if (isNumber(viewName)) { 
      index = Number(viewName);
    } else {
      index = getViewIndexByViewName(views, viewName);
    }
    return index;
  }
  
  function isAppPath(url) {
    return url.substring(0, 11) === 'app-path://';
  }
  
  function removeAppPathProtocol(url) {
    return url.substring(11, url.length);
  }
  
  function getViewByAppPath(url) {
    var basePath,
        pathTokens,
        viewName,
        parentOfViews,
        viewIndex;
        
    basePath = removeAppPathProtocol(url); //Remove protocol (i.e. app-path://) from url 
    pathTokens = basePath.split('/');
    viewName = pathTokens[pathTokens.length - 1]; 
    parentOfViews = getParentObjOfViewArray(pathTokens);
    viewIndex = getViewIndex(parentOfViews.views, viewName);
    return parentOfViews.views[viewIndex];
  }
  
  function parseViewProperty(view, propertyName) {
    var pathTokens,
        basePath,
        viewPropertyValue,
        url;
    if (!view[propertyName]) {
      throw 'Invalid view object. Missing property "' + propertyName + '"!';
    } else {
      url = view[propertyName].url;
      if (url) {
        if (isAppPath(url)) {
          view = getViewByAppPath(url);
          viewPropertyValue = parseViewProperty(view, 'template'); 
        } else {
          viewPropertyValue = app.xhr.ajax(url);
        }
      } else {
        if (view[propertyName].value) {
          viewPropertyValue = view[propertyName].value;
        } 
      }
    }
    
    if (viewPropertyValue &&
        propertyName === 'data' &&
        typeof viewPropertyValue === 'string') {
        viewPropertyValue = JSON.parse(viewPropertyValue);
    }
    
    view[propertyName].value = viewPropertyValue || '';
    return view[propertyName].value;
  }
  
  function getViewData(view) {
    return parseViewProperty(view, 'data');
  }
  
  function getViewInitMethod(view) {
    var pathTokens,
        methodName,
        viewInitRefObj;
    if (view.init) {
      pathTokens = view.init.split('/');
      methodName = pathTokens[pathTokens.length - 1];
      viewInitRefObj = getParentObjOfViewArray(pathTokens);
      return viewInitRefObj[methodName]; //return the function value; not invoking it here
    }
  }
  
  function getViewTemplate(view) {
    return parseViewProperty(view, 'template'); 
  }
  
  function renderInContainer(containerId, htmlTemplate, data, init) {
      var html = app.Mustache.to_html(htmlTemplate, data);
      document.getElementById(containerId).innerHTML = html;
      if (init) {
        init();
      }
  }
  
  function renderView(view) {
    var containerId = view.container_id,
        viewData = getViewData(view),
        viewTemplate = getViewTemplate(view),
        init = getViewInitMethod(view);        
    renderInContainer(containerId, viewTemplate, viewData, init);
  }
  
  function getViewByUrl(url) {
    return JSON.parse(app.xhr.ajax(url));
  }
  
  function renderViewByUrl(url) {
    var view;
        
    if (isAppPath(url)) {
      view = getViewByAppPath(url);
    } else {
      view = getViewByUrl(url);
    }
    renderView(view);
  }
  
  function renderViewByViewObj(view) {
    renderView(view);
  }
  
  /** engine API **/
	engine = {
    name: "engine.js",
    version: "0.4.1",
    render: function (view) {
      if (typeof view === 'string') {
        renderViewByUrl(view);
      } else if (typeof view  === 'object') {
        renderViewByViewObj(view);
      } else {
        throw 'Invalid view type, expected string or object. view: ' + JSON.stringify(view);
      }
    },
    Mustache: {
      to_html: function () {
        throw 'missing Mustache plugin';
      }
    },
    xhr: (function (window) {
      var xmlhttp;
          
      function init(settings) {
        if (window.XMLHttpRequest) {// code for ie7+, firefox, chrome, opera, safari
          xmlhttp = new XMLHttpRequest();
        } else {// code for ie6, ie5
          xmlhttp = new ActiveXObject('Microsoft.XMLHTTP');
        }
        
        xmlhttp.onreadystatechange = function (settings) {
          if (xmlhttp.readystate === 4 && xmlhttp.status === 200) {
            if (settings.success) {
              settings.success(xmlhttp.responseText);
            }
          }
        };
      }
      
      function open (url, settings) {
        xmlhttp.open(settings.type, url, settings.async);
      }
      
      return {
        name: "xhr.js",
        version: "0.1.0",
        settings: {
          async: false,
          type: 'GET',
          success: null,
          error: null
        },
        ajax: function(url, overrideSettings) {
          var response = '',
              settings;
          try {
            if (!xmlhttp) {
              init();
            }
            settings = overrideSettings || this.settings;
            open(url, settings);
            xmlhttp.send();
            if (!settings.async) {
              if (xmlhttp.status === 200) {
                response = xmlhttp.response;
              } else {
                response = '';
              }
            }
          } catch(e) {
            if (settings.error) {
              settings.error(e);
            }
          }
          return response;
        }
      };
    }(window))
  };
	
  return engine;
}());
